"""Tracker rules processor mixin."""

from types import *
import time, whrandom

import Acquisition
from Persistence import Persistent

DEBUG = 0

RULES_USER = '.rules.'

class TransactionAbort(Exception):
    pass

BadForm = "BadForm"           # String exception, needs to be caught elsewhere.

class Noodge(Acquisition.Implicit,
             Persistent,
             ):
    """Maintain and process the tracker's rule base.

    Rules are applied only to active issues - ie, those that are not
    in the completed, rejected, or deferred stage."""
    def __init__(self, rules=None):
        self.init_expr_tables()
        self.set_rules(rules)

    def set_rules(self, rules):
        self._rules = rules
    def init_expr_tables(self):
        # Map non-builtin conditions to method names.
        # XXX Eventually this will 
        self._conditions = {'state': "state_condition",
                            }
        self._actions = {'notify': "notify_action",
                         }

    def execute_rules(self, issue=None):
        """Apply rules to all issues, or specified one, if any.

        Return value is a list of exceptions encountered - tuples of the
        exception, details, and rule - if any.

        NOTE BENE - invoking execute_rules within an ongoing transaction
        for an issue will wipe the transaction clean up to the point of
        applying the rules.  The state ticker for each issue is incremented 
        after rules are executed for it, marking all surviving issue
        audited state as no longer volatile."""
        rules = self._rules
        if not rules:
            return
        if issue is not None:
            issues = [issue]
            single_issue = 1
        else:
            issues = self._issues.values()
            single_issue = 0
        todo = []

        exceptions = []

        for i in issues:
            if (i.audited_state.get('stage', None)
                in ['completed', 'rejected', 'deferred']):
                continue

            transaction = None
            try:
                transaction = get_transaction()
                if not single_issue:
                    transaction.begin()
            except:
                transaction = None

            aborted = 0
            try:
                for rule in self._rules:
                    self.eval_rule(rule, i)
            except TransactionAbort, details:
                if transaction:
                    transaction.abort()
                aborted = 1
                exceptions.append(TransactionAbort, details, rule)
                if DEBUG >= 2:
                    Note("TransactionAbort(%s) encountered for %s\n",
                         details, `rule`)
            if not aborted and transaction and not single_issue:
                transaction.commit()
            if DEBUG >= 1:
                i.audited_state.set('ranrules', i.tick(), RULES_USER)
            i.increment_tick()
                
        return exceptions

    def eval_rule(self, rule, issue):
        """Evaluate a rule, returning qualified actions.

        Returns 1 if any actions were executed, else 0 - unless:

        Raises TransactionAbort if an 'abort' action is executed.

        Raises BadForm if invalid rule is encountered."""

        # Rules are lists of data structures called contingencies:
        #
        #  [rule, ...] 
        # 
        # a rule contains 0 or more conditions and 1 or more
        # actions.
        #
        # <rule> :   ( ('condition', <condition>), <rule>, ... )
        #          | ('actions' (<1 or more actions>))
        #
        # A leading ('condition', <condition>), if any, must hold for any
        # of the subcontingencies to be evaluated.

        try:
            selector = rule[0]
        except (AttributeError, IndexError):
            return rule          # A literal.

        if selector == 'actions':
            return self.apply_actions(issue, rule[1:])
        elif type(selector) != TupleType:
            raise BadForm, ("Composite rule needs leading condition: %s"
                            % `rule`)
        elif selector[0] != 'condition':
            raise BadForm, ("Malformed Rule condition: %s"
                            % `rule`)
        else:
            did = 0
            if self.eval_condition(selector[1], issue):
                for i in rule[1:]:
                    did = self.eval_rule(i, issue) or did
            return did
        
    ###### Conditions ######

    def eval_condition(self, condition, issue):
        """Evaluate w.r.t issue and return truth or time (secs) val."""

        # condition is either a literal (number, etc - precisely, raises
        # AttributeError or IndexError upon index lookup) or a tuple
        # expression, either one of the following builtins or one defined
        # (according to its first-element selector) in the._conditions
        # dispatch table.
        # 
        #  ('and' <condition> ...)	  - logical conjunction
        #  ('or' <condition> ...)	  - logical disjunction
        #  ('not' <condition> ...)	  - logical negation
        #  ('>' <condition> <condition>)  - Greater than
        #  ('<' <condition> <condition>)  - Less than
        #  ('=' <condition> ...)	  - equals
        #  ('contains' <state-var> <value(s)>) - state var includes value
        #  ('volatile' <state-var> ...)	  - var got new value this tick

        if type(condition) == StringType:
            raise BadForm, ("Invalid condition - string: %s" % `condition`)

        try:
            cname = condition[0]
        except (AttributeError, IndexError, TypeError):
            return condition            # A literal.
        except TypeError, msg:
            raise TypeError, (msg, `condition`)

        if cname == 'and':
            for i in condition[1:]:
                if not self.eval_condition(i, issue):
                    return 0
            return 1
        elif cname == 'or':
            for i in condition[1:]:
                if self.eval_condition(i, issue):
                    return 1
            return 0
        elif cname == 'not':
            if len(condition) != 2:
                raise BadForm, ("Wrong number of args, need 1, got %d: %s"
                                % (len(condition)-1, `condition`))
            return not self.eval_condition(condition[1], issue)
        elif cname == '<':
            last = None
            for i in condition[1:]:
                next = self.eval_condition(i, issue)
                if last is not None and last >= next:
                    return 0
                last = next
            return 1
        elif cname == '>':
            last = None
            for i in condition[1:]:
                next = self.eval_condition(i, issue)
                if last is not None and last <= next:
                    return 0
                last = next
            return 1
        elif cname == '=':
            last = None
            for i in condition[1:]:
                next = self.eval_condition(i, issue)
                if last is None:
                    last = next
                elif last != next:
                    return 0
            return 1
        elif cname == 'contains':
            if len(condition) != 3:
                raise BadForm, ("Wrong number of args, need 2, got %d: %s"
                                % (len(condition)-1, `condition`))
            values = issue.audited_state.get(condition[1])
            if type(values) != DictType:
                values = {values: None}
            return values.has_key(condition[2])
        elif cname == 'volatile':
            history = issue.audited_state.history
            tick = issue.tick()
            for i in condition[1:]:
                if not history(i, tick=tick):
                    return 0
            return 1
        else:
            # Get from the table.
            methnm = self._conditions.get(cname, None)
            if not methnm:
                raise BadForm, ("Unrecognized condition '%s'"
                                % cname)
            elif not hasattr(self, methnm):
                raise SystemError, ("Improperly registered condition '%s'"
                                    % cname)
            meth = getattr(self, methnm)
            try:
                return apply(meth, (issue,) + tuple(condition[1:]))
            except (BadForm, TypeError), err:
                raise BadForm, str(err) + ": " + `condition`

    def state_condition(self, issue, var=None, value=None, priorcount=None):
        """Return # of seconds since state var was set to value, or 0 if not.
        
        Many forms:

         ('state') returns the "age" of the issue - seconds since creation.

         ('state' 'var') or ('state' 'var' None 0)
                   returns duration since var's most recent setting, or 0.

         ('state' 'var' val)
                   returns duration that varname has had current setting
                   equal to value, or 0 if it doesn't have that value.
                                 
         ('state' 'var' None n>0)
                   returns time since var got its n'th prior setting.

         ('state' 'var' val 0)
                   returns time since varname was last set to val, or 0.

         ('state' 'var' val n>0)
                   returns time since var got its n'th prior setting to
                   val, or 0.
        
        When value is a list, a value match for any of the items on the
        list satisfies."""

        state = issue.audited_state

        if not var:                # ('state')
            if value is not None:
                raise BadForm, "Invalid: value without state var"

            return state.age()

        x = ['unique']
        if state.get(var, x) is x:
            return 0                    # no setting for var

        # Assert: var has at least one setting.

        now = time.time()

        if value is None:
            if priorcount in [None, 0]:
                # ('state', "var") or ('state', "var", None, 0)
                return now - state.history(varname=var, max=1)[0][-1]
            else:                       # ('state', "var", None, X):
                got = state.history(varname=var, max=priorcount+1)
                if len(got) < (priorcount+1):
                    return 0
                else:
                    return now - got[-1][-1]  
        else:

            if type(value) != ListType:
                value = [value]

            if priorcount is None:      # ('state', "var", value):
                curval = state.get(var)
                if curval not in value:
                    return 0
                return now - state.history(varname=var, max=1)[0][-2]

            elif priorcount == 0:       # ('state', "var", value, 0):
                got = state.history(varname=var, value=value, max=1)
                if not got:
                    return 0
                else:
                    return now - got[0][-2]

            elif priorcount > 0:        # ('state', "var", value, n):
                got = state.history(varname=var,
                                    value=value, max=1+priorcount)
                if len(got) < (priorcount):
                    return 0
                else:
                    return now - got[-1][-2]

            else:
                raise BadForm, "Invalid 'state' condition priorcount"

    ###### Actions ######

    def apply_actions(self, issue, actions):
        """Invoke the specified actions in the context of the issue."""

        #  ('set' <state-var> <value>)      - Set state var to value
        #  ('include' <state-var> <value>)  - State var gets value among others
        #  ('remove' <state-var> <value>)   - State var no longer has value
        #  ('notify' <role> <subject> <msg>) - Send a notice, text is dtml
        #  ('abort' <msg>)		    - Forbidden - rollback transaction

        state = issue.audited_state
        for a in actions:
            try:
                selector = a[0]
            except (AttributeError, IndexError), err:
                raise BadForm, ("Invalid action - string: %s" % `a`)

            if selector == 'set':
                if len(a) != 3:
                    raise BadForm, ("Invalid 'set' action - "
                                    "wrong # of args: %s" % `a`)
                state.set(a[1], a[2], RULES_USER)
            elif selector == 'include':
                if len(a) != 3:
                    raise BadForm, ("Invalid 'include' action - "
                                    "wrong # of args: %s" % `a`)
                curval = state.get(a[1], {})
                if type(curval) is not DictType:
                    # Yes, we automatically promote to being a bag.
                    curval = {curval: None}
                newval = {}; newval.update(curval)
                newval[a[2]] = None
                if newval != curval:
                    state.set(a[1], newval, RULES_USER)
            elif selector == 'remove':
                if len(a) != 3:
                    raise BadForm, ("Invalid 'remove' action - "
                                    "wrong # of args: %s" % `a`)
                curval = state.get(a[1], {})
                if type(curval) is not DictType:
                    # Yes, we automatically promote to being a bag.
                    curval = {curval: None}
                newval = {}; newval.update(curval)
                del newval[a[2]]
                if newval != curval:
                    state.set(a[1], newval, RULES_USER)
            elif selector == 'abort':
                errargs = tuple(a[1:])
                if not errargs: errargs = None
                raise TransactionAbort, errargs
            else:
                # Get from the table.
                methnm = self._actions.get(a[0], None)
                if not methnm:
                    raise BadForm, ("Unrecognized action '%s'"
                                    % a[0])
                elif not hasattr(self, methnm):
                    raise SystemError, ("Improperly registered action '%s'"
                                        % a[0])
                meth = getattr(self, methnm)
                try:
                    apply(meth, (issue, RULES_USER) + tuple(a[1:]))
                except (BadForm, TypeError), err:
                    raise BadForm, str(err) + ": " + `a`
        
        return 1

    def notify_action(self, issue, user, who, subj, msg):
        # *** set 'msgfromrole' history var
        return self._notifier.notify(issue, who, subj, msg) or 1

def Note(text, *args):
    import sys
    try:
        msg = text % args
    except TypeError:
        msg = "Note - faulty note: %s, %s" % (`text`, `args`)
    __traceback_info__ = ("Note:", msg)
    try:
        sys.stderr.write(msg)
    except IOError:
        pass
